div.nbinput.container,
div.nboutput.container {
    /* Order flex items (toggle button and contents) in one column */
    flex-direction: column;
}

div.nbinput.container * {
    /* We must set the width to a 100% */
    width: 100%;
}

.nbinput .prompt,
.nboutput .prompt {
    /* Remove prompts, otherwise the layout is messed up */
    display: none;
}

.my-nbsphinx-showbutton,
.my-nbsphinx-hidebutton {
    cursor: pointer;
	padding: 5px;
    font-size: 12px;
    font-weight: bold;
	font-family: monospace;
	text-align: center;
}

.my-nbsphinx-showbutton {
    /* Block is folded, hide the showbutton */
    display: none;
}
.my-nbsphinx-folded .my-nbsphinx-showbutton {
    /* Block is folded, show the showbutton */
    display: block;
}
.my-nbsphinx-folded .my-nbsphinx-hidebutton {
    /* Block is folded, hide the hidebutton */
    display: none;
}

.my-nbsphinx-folded .input_area,
.my-nbsphinx-folded .output_area {
    /* Hide the contents then the parent has the my-nbsphinx-folded class */
    display: none;
}
