@import url("css/theme.css");

html[data-theme="light"] {
    --pst-color-primary: #5d7980;     /* dark grey */
    --pst-color-secondary: #e86B35;   /* dark version of calorine orange */
    --pst-color-background: #fcfcfc;  /* off-white */
    --pst-color-accent: #3bdaff;      /* light blue */
    --pst-color-info: #ff763b;        /* calorine orange */
    --pst-color-info-bg: #ffede6;     /* calorine orange */
}

html[data-theme="dark"] {
    --pst-color-primary: #ff9466;        /* orange */
    --pst-color-secondary: #ffb190;      /* light orange */
    --pst-color-background: #14181e;     /* dark gray */
    --pst-color-accent: #ff763b;         /* calorine orange */
    --pst-color-info: #ff9466;           /* same as primary */
    --pst-color-info-bg: #612916;        /* dark orange */
    --pst-color-inline-code-links: #ff9466;
}

/* Add FA external link icon after all external links */
a.reference.external::after {
  font-family: "Font Awesome 6 Free";
  font-weight: 900;
  content: "\f35d";
  margin-left: 0.3em;
  font-size: 0.85em;
  vertical-align: middle;
  color: #999;
  opacity: 0.7;
}
